@@ -303,7 +303,7 @@ int fw_iso_recv_start(fw_handle_t handle, int start_on_cycle,
 }
 
 static int handle_iso_event(raw1394handle_t handle,
-			    struct epoll_closure *closure, __uint32_t events)
+			    struct epoll_closure *closure, uint32_t events)
 {
 	fw_handle_t fwhandle = handle->mode.fw;
 	struct fw_cdev_event_iso_interrupt *interrupt;
@@ -322,10 +322,9 @@ static int handle_iso_event(raw1394handle_t handle,
 	{
 		int cycle;
 
-		fwhandle->iso.packet_count -= fwhandle->iso.irq_interval;
-
 		/* Check whether the ABI version provides iso tx timestamps. */
 		if (interrupt->header_length) {
+			fwhandle->iso.packet_count -= interrupt->header_length/4;
 			/*
 			 * Take the cycle of the last packet transmitted, add
 			 * the number of packets currently queued, plus one, and
@@ -335,6 +334,7 @@ static int handle_iso_event(raw1394handle_t handle,
 			cycle = be32_to_cpu(interrupt->header[interrupt->header_length/4 - 1]);
 			cycle &= 0x1fff;
 		} else {
+			fwhandle->iso.packet_count -= fwhandle->iso.irq_interval;
 			/*
 			 * Bogusly faking it again.  Assume that the last packet
 			 * transmitted was transmitted on interrupt->cycle.
@@ -431,21 +431,10 @@ int fw_iso_xmit_sync(raw1394handle_t handle)
 
 int fw_iso_recv_flush(fw_handle_t handle)
 {
-	/* FIXME: huh, we'll need kernel support here... */
+	struct fw_cdev_flush_iso flush;
 
-	return 0;
-}
-
-static unsigned int
-round_to_power_of_two(unsigned int value)
-{
-	unsigned int pot;
-
-	pot = 1;
-	while (pot < value)
-		pot <<= 1;
-
-	return pot;
+	flush.handle = handle->iso.kernel_handle;
+	return ioctl(handle->iso.fd, FW_CDEV_IOC_FLUSH_ISO, &flush);
 }
 
 static int
@@ -460,25 +449,13 @@ iso_init(fw_handle_t handle, int type,
 {
 	struct fw_cdev_create_iso_context create;
 	struct epoll_event ep;
-	int retval, prot;
+	int retval;
 
 	if (handle->iso.fd != -1) {
 		errno = EBUSY;
 		return -1;
 	}
 
-	switch (type) {
-	case FW_CDEV_ISO_CONTEXT_TRANSMIT:
-		prot = PROT_READ | PROT_WRITE;
-		break;
-	case FW_CDEV_ISO_CONTEXT_RECEIVE:
-		prot = PROT_READ;
-		break;
-	default:
-		errno = EINVAL;
-		return -1;
-	}
-
 	/* set irq_interval from < 1 to default values like ieee1394 rawiso */
 	if (irq_interval < 0)
 		irq_interval = buf_packets / 4;
@@ -490,7 +467,7 @@ iso_init(fw_handle_t handle, int type,
 	handle->iso.xmit_handler = xmit_handler;
 	handle->iso.recv_handler = recv_handler;
 	handle->iso.buf_packets = buf_packets;
-	handle->iso.max_packet_size = round_to_power_of_two(max_packet_size);
+	handle->iso.max_packet_size = max_packet_size;
 	handle->iso.packet_phase = 0;
 	handle->iso.packet_count = 0;
 	handle->iso.packets =
@@ -537,7 +514,7 @@ iso_init(fw_handle_t handle, int type,
 
 	handle->iso.buffer =
 		mmap(NULL, buf_packets * handle->iso.max_packet_size,
-		     prot, MAP_SHARED, handle->iso.fd, 0);
+		     PROT_READ | PROT_WRITE, MAP_SHARED, handle->iso.fd, 0);
 
 	if (handle->iso.buffer == MAP_FAILED) {
 		close(handle->iso.fd);
@@ -647,10 +624,9 @@ void fw_iso_shutdown(fw_handle_t handle)
 	handle->iso.fd = -1;
 }
 
-int fw_read_cycle_timer(fw_handle_t handle,
-			u_int32_t *cycle_timer, u_int64_t *local_time)
+int fw_read_cycle_timer(fw_handle_t handle, u_int32_t *cycle_timer,
+			  u_int64_t *local_time)
 {
-#ifdef FW_CDEV_IOC_GET_CYCLE_TIMER /* added in kernel 2.6.24 */
 	int err;
 	struct fw_cdev_get_cycle_timer ctr = { 0 };
 
@@ -660,8 +636,19 @@ int fw_read_cycle_timer(fw_handle_t handle,
 		*local_time  = ctr.local_time;
 	}
 	return err;
-#else
-	errno = ENOSYS;
-	return -1;
-#endif /* defined(FW_CDEV_IOC_GET_CYCLE_TIMER) */
+}
+
+int fw_read_cycle_timer_and_clock(fw_handle_t handle, u_int32_t *cycle_timer,
+			  u_int64_t *local_time, clockid_t clk_id)
+{
+	int err;
+	struct fw_cdev_get_cycle_timer2 ctr = {.clk_id = clk_id};
+
+	err = ioctl(handle->ioctl_fd, FW_CDEV_IOC_GET_CYCLE_TIMER2, &ctr);
+	if (!err) {
+		*cycle_timer = ctr.cycle_timer;
+		*local_time  = ctr.tv_sec * 1000000ULL +
+			       ctr.tv_nsec / 1000;
+	}
+	return err;
 }
