@@ -17,6 +17,7 @@
 #include <sys/ioctl.h>
 #include <sys/epoll.h>
 #include <sys/inotify.h>
+#include <time.h>
 #include <unistd.h>
 #include <fcntl.h>
 #include <errno.h>
@@ -26,13 +27,9 @@
 #include "raw1394_private.h"
 
 /*
- * ABI version history is documented in linux/firewire-cdev.h.
+ * ABI version history is documented in firewire-cdev.h.
  */
-#ifdef FW_CDEV_EVENT_REQUEST2
 #define IMPLEMENTED_CDEV_ABI_VERSION	4
-#else
-#define IMPLEMENTED_CDEV_ABI_VERSION	2
-#endif
 
 int
 fw_errcode_to_errno(raw1394_errcode_t errcode)
@@ -190,7 +187,7 @@ scan_devices(fw_handle_t handle)
 
 static int
 handle_echo_pipe(raw1394handle_t handle,
-		 struct epoll_closure *ec, __uint32_t events)
+		 struct epoll_closure *ec, uint32_t events)
 {
 	quadlet_t value;
 
@@ -261,7 +258,7 @@ handle_fcp_request(raw1394handle_t handle, struct address_closure *ac,
 
 static int
 handle_device_event(raw1394handle_t handle,
-		    struct epoll_closure *ec, __uint32_t events)
+		    struct epoll_closure *ec, uint32_t events)
 {
 	fw_handle_t fwhandle = handle->mode.fw;
 	union fw_cdev_event *u;
@@ -323,7 +320,6 @@ handle_device_event(raw1394handle_t handle,
 				    u->request.handle,
 				    u->request.length, u->request.data);
 
-#ifdef FW_CDEV_EVENT_REQUEST2 /* added in kernel 2.6.36 */
 	case FW_CDEV_EVENT_REQUEST2:
 		ac = u64_to_ptr(u->request.closure);
 		return ac->callback(handle, ac, u->request2.tcode,
@@ -332,17 +328,13 @@ handle_device_event(raw1394handle_t handle,
 				    u->request2.card,
 				    u->request2.handle,
 				    u->request2.length, u->request2.data);
-#endif
 
-#ifdef FW_CDEV_EVENT_ISO_RESOURCE_ALLOCATED /* added in kernel 2.6.30 */
 	case FW_CDEV_EVENT_ISO_RESOURCE_ALLOCATED:
 	case FW_CDEV_EVENT_ISO_RESOURCE_DEALLOCATED:
 		memcpy(u64_to_ptr(u->iso_resource.closure), u,
 		       sizeof u->iso_resource);
 		return 0;
-#endif
 
-#ifdef FW_CDEV_EVENT_PHY_PACKET_SENT /* added in kernel 2.6.36 */
 	case FW_CDEV_EVENT_PHY_PACKET_SENT:
 		rc = u64_to_ptr(u->phy_packet.closure);
 		errcode = fw_to_raw1394_errcode(u->phy_packet.rcode);
@@ -350,7 +342,7 @@ handle_device_event(raw1394handle_t handle,
 		free(rc);
 
 		return fwhandle->tag_handler(handle, tag, errcode);
-#endif
+
 	default:
 	case FW_CDEV_EVENT_ISO_INTERRUPT:
 		/* Never happens. */
@@ -359,21 +351,16 @@ handle_device_event(raw1394handle_t handle,
 }
 
 static int
-handle_inotify(raw1394handle_t handle, struct epoll_closure *ec,
-	       __uint32_t events)
+process_inotify_event(fw_handle_t fwhandle, struct inotify_event *event)
 {
-	fw_handle_t fwhandle = handle->mode.fw;
-	struct inotify_event *event;
 	char filename[32];
 	struct fw_cdev_get_info info;
 	struct fw_cdev_event_bus_reset reset;
 	struct epoll_event ep;
-	int i, len, fd, phy_id, fname_str_sz;
+	int i, fd, phy_id, fname_str_sz;
 
-	event = (struct inotify_event *) fwhandle->buffer;
-	len = read(fwhandle->inotify_fd, event, BUFFER_SIZE);
 	if (!(event->mask & IN_CREATE))
-		return -1;
+		return 0;
 	if (!is_fw_device_name(event->name))
 		return 0;
 	snprintf(filename, sizeof filename, FW_DEVICE_DIR "/%s", event->name);
@@ -425,12 +412,35 @@ handle_inotify(raw1394handle_t handle, struct epoll_closure *ec,
 	ep.data.ptr = &fwhandle->devices[i].closure;
 	if (epoll_ctl(fwhandle->epoll_fd, EPOLL_CTL_ADD, fd, &ep) < 0) {
 		close(fd);
+		fwhandle->devices[i].fd = -1;
 		return -1;
 	}
 
 	return 0;
 }
 
+static int
+handle_inotify(raw1394handle_t handle, struct epoll_closure *ec,
+	       uint32_t events)
+{
+	fw_handle_t fwhandle = handle->mode.fw;
+	struct inotify_event *event;
+	ssize_t len;
+	int retval = 0;
+
+	event = (struct inotify_event *) fwhandle->buffer;
+	len = read(fwhandle->inotify_fd, event, BUFFER_SIZE);
+
+	while (len >= sizeof(struct inotify_event)) {
+		retval |= process_inotify_event(fwhandle, event);
+		len -= sizeof(struct inotify_event) + event->len;
+		event = (struct inotify_event *) ((char *)event +
+				(sizeof(struct inotify_event) + event->len));
+	}
+
+	return retval;
+}
+
 int fw_loop_iterate(raw1394handle_t handle)
 {
 	int i, count, retval = 0;
@@ -592,6 +602,34 @@ int fw_get_nodecount(fw_handle_t handle)
 	return (handle->reset.root_node_id & 0x3f) + 1;
 }
 
+int fw_get_speed(fw_handle_t handle, nodeid_t node)
+{
+	int i;
+
+	if ((node & ~0x3f) != 0xffc0) {
+		errno = ENOSYS;
+		return -1;
+	}
+
+	if (node > handle->reset.root_node_id) {
+		errno = fw_errcode_to_errno(-RCODE_NO_ACK);
+		return -1;
+	}
+
+	i = handle->nodes[node & 0x3f];
+	if (i == -1) {
+		errno = fw_errcode_to_errno(-RCODE_NO_ACK);
+		return -1;
+	}
+
+	if (handle->generation != handle->devices[i].generation) {
+		errno = fw_errcode_to_errno(-RCODE_GENERATION);
+		return -1;
+	}
+
+	return ioctl(handle->devices[i].fd, FW_CDEV_IOC_GET_SPEED);
+}
+
 int fw_get_port_info(fw_handle_t handle,
 			  struct raw1394_portinfo *pinf,
 			  int maxports)
@@ -755,7 +793,7 @@ handle_arm_request(raw1394handle_t handle, struct address_closure *ac,
 	struct fw_cdev_send_response response;
 	arm_options_t type;
 	size_t in_length;
-	int pos;
+	int pos, retval;
 
 	pos = offset - allocation->offset;
 	response.handle = kernel_handle;
@@ -832,7 +870,7 @@ handle_arm_request(raw1394handle_t handle, struct address_closure *ac,
 
 	rrb = malloc(sizeof *rrb + in_length + response.length);
 	if (rrb == NULL) {
-		errno == ENOMEM;
+		errno = ENOMEM;
 		return -1;
 	}
 
@@ -860,8 +898,10 @@ handle_arm_request(raw1394handle_t handle, struct address_closure *ac,
 	rrb->response.buffer = rrb->data + in_length;
 	memcpy(rrb->response.buffer, allocation->data + pos, response.length);
 
-	return fwhandle->arm_tag_handler(handle, allocation->tag, type,
-					 length, &rrb->request_response);
+	retval = fwhandle->arm_tag_handler(handle, allocation->tag, type,
+					   length, &rrb->request_response);
+	free(rrb);
+	return retval;
 }
 
 int
@@ -895,9 +935,8 @@ fw_arm_register(fw_handle_t handle, nodeaddr_t start,
 	request.offset = start;
 	request.length = length;
 	request.closure = ptr_to_u64(&allocation->closure);
-#if IMPLEMENTED_CDEV_ABI_VERSION >= 4
 	request.region_end = start + length;
-#endif
+
 	retval = ioctl(handle->ioctl_fd, FW_CDEV_IOC_ALLOCATE, &request);
 	if (retval < 0) {
 		free(allocation);
@@ -1003,24 +1042,16 @@ send_request(fw_handle_t handle, int tcode,
 	int ioctl_nr = FW_CDEV_IOC_SEND_REQUEST;
 	int fd, i, retval;
 
-#ifdef FW_CDEV_IOC_SEND_STREAM_PACKET /* added in kernel 2.6.30 */
 	if (tcode == TCODE_STREAM_DATA) {
 		ioctl_nr = FW_CDEV_IOC_SEND_STREAM_PACKET;
 		fd = handle->ioctl_fd;
 	}
-#else
-	if (tcode == TCODE_STREAM_DATA) {
-		errno = ENOSYS;
-		return -1;
-	}
-#endif
 
-#ifdef FW_CDEV_IOC_SEND_BROADCAST_REQUEST /* added in kernel 2.6.30 */
 	if (node == 0xffff) {
 		ioctl_nr = FW_CDEV_IOC_SEND_BROADCAST_REQUEST;
 		fd = handle->ioctl_fd;
 	}
-#endif
+
 	if (ioctl_nr != FW_CDEV_IOC_SEND_REQUEST)
 		goto node_id_ok;
 
@@ -1064,7 +1095,6 @@ node_id_ok:
 	request->data       = ptr_to_u64(in);
 	request->generation = handle->generation;
 
-#ifdef FW_CDEV_IOC_SEND_STREAM_PACKET
 	if (tcode == TCODE_STREAM_DATA) {
 		struct fw_cdev_send_stream_packet *p
 		    = (struct fw_cdev_send_stream_packet *) request;
@@ -1078,7 +1108,6 @@ node_id_ok:
 		p->generation = handle->generation;
 		p->speed      = (addr >> 4) & 0x7;
 	}
-#endif
 
 	retval = ioctl(fd, ioctl_nr, request);
 	if (retval < 0)
@@ -1215,7 +1244,6 @@ fw_start_async_stream(fw_handle_t handle, unsigned int channel,
 int
 fw_start_phy_packet_write(fw_handle_t handle, quadlet_t data, unsigned long tag)
 {
-#ifdef FW_CDEV_IOC_SEND_PHY_PACKET /* added in kernel 2.6.36 */
 	struct fw_cdev_send_phy_packet send_phy_packet;
 	struct request_closure *closure;
 	int retval;
@@ -1246,11 +1274,6 @@ fw_start_phy_packet_write(fw_handle_t handle, quadlet_t data, unsigned long tag)
 		free(closure);
 
 	return retval;
-#else
-	errno = ENOSYS;
-	handle->err = -errno;
-	return -1;
-#endif
 }
 
 int
@@ -1281,7 +1304,7 @@ sync_callback(raw1394handle_t handle, void *data, raw1394_errcode_t err)
 }
 
 static int
-send_request_sync(raw1394handle_t handle, int tcode,
+_send_request_sync(raw1394handle_t handle, int tcode,
 		  nodeid_t node, nodeaddr_t addr,
 		  size_t in_length, void *in, size_t out_length, void *out)
 {
@@ -1309,12 +1332,91 @@ send_request_sync(raw1394handle_t handle, int tcode,
 	return (errno ? -1 : 0);
 }
 
+static int
+send_request_sync(raw1394handle_t handle, int tcode,
+		  nodeid_t node, nodeaddr_t addr,
+		  size_t in_length, void *in, size_t out_length, void *out)
+{
+	static const struct timespec delay = {
+		.tv_sec  = 0,
+		.tv_nsec = 10 * 1000 * 1000,
+	};
+	fw_handle_t fwhandle = handle->mode.fw;
+	int i, ret;
+
+	/*
+	 * Retry after ack-busy, but give the responder some breathing room.
+	 * 10 ms delay between retries is proper for ancient camcorders.
+	 */
+	for (i = 1; ; i++) {
+		ret = _send_request_sync(handle, tcode, node, addr,
+					 in_length, in, out_length, out);
+		if (ret == 0 || i == 10 || fwhandle->err != -RCODE_BUSY)
+			break;
+		nanosleep(&delay, NULL);
+	}
+
+	return ret;
+}
+
+static int
+read_config_rom(fw_handle_t handle, nodeid_t node, nodeaddr_t addr,
+	     size_t length, quadlet_t *buffer)
+{
+	struct fw_cdev_get_info get_info;
+	quadlet_t rom[256];
+	int offset, fd, i, err;
+
+	if (node > handle->reset.root_node_id) {
+		handle->err = -RCODE_NO_ACK;
+		errno = fw_errcode_to_errno(handle->err);
+		return -1;
+	}
+
+	i = handle->nodes[node & 0x3f];
+	if (i == -1) {
+		handle->err = -RCODE_NO_ACK;
+		errno = fw_errcode_to_errno(handle->err);
+		return -1;
+	}
+
+	if (handle->generation != handle->devices[i].generation) {
+		handle->err = -RCODE_GENERATION;
+		errno = fw_errcode_to_errno(handle->err);
+		return -1;
+	}
+	fd = handle->devices[i].fd;
+
+	memset(&get_info, 0, sizeof(get_info));
+	memset(&rom,      0, sizeof(rom));
+	get_info.version    = IMPLEMENTED_CDEV_ABI_VERSION;
+	get_info.rom        = ptr_to_u64(rom);
+	get_info.rom_length = sizeof(rom);
+	get_info.bus_reset  = 0;
+
+	err = ioctl(fd, FW_CDEV_IOC_GET_INFO, &get_info);
+	if (err)
+		return err;
+
+	offset = (addr - CSR_REGISTER_BASE - CSR_CONFIG_ROM) / 4;
+	for (i = 0; i < length / 4; i++)
+		buffer[i] = cpu_to_be32(rom[offset + i]);
+
+	return 0;
+}
+
 int
 fw_read(raw1394handle_t handle, nodeid_t node, nodeaddr_t addr,
 	     size_t length, quadlet_t *buffer)
 {
 	int tcode;
 
+	if (addr >= CSR_REGISTER_BASE + CSR_CONFIG_ROM &&
+	    addr + length <= CSR_REGISTER_BASE + CSR_CONFIG_ROM_END &&
+	    !(addr & 3) && length > 0 && !(length & 3))
+		return read_config_rom(handle->mode.fw,
+				       node, addr, length, buffer);
+
 	if (length == 4)
 		tcode = TCODE_READ_QUADLET_REQUEST;
 	else
@@ -1442,9 +1544,8 @@ fw_start_fcp_listen(fw_handle_t handle)
 	request.offset = CSR_REGISTER_BASE + CSR_FCP_COMMAND;
 	request.length = CSR_FCP_END - CSR_FCP_COMMAND;
 	request.closure = ptr_to_u64(closure);
-#if IMPLEMENTED_CDEV_ABI_VERSION >= 4
 	request.region_end = CSR_REGISTER_BASE + CSR_FCP_END;
-#endif
+
 	if (ioctl(handle->ioctl_fd, FW_CDEV_IOC_ALLOCATE, &request) < 0)
 		return -1;
 
@@ -1471,6 +1572,40 @@ fw_update_config_rom(fw_handle_t handle, const quadlet_t *new_rom,
 	return -1;
 }
 
+int
+fw_add_config_rom_descriptor(fw_handle_t handle, u_int32_t *token,
+		quadlet_t immediate_key, quadlet_t key,
+		const quadlet_t *data, size_t size)
+{
+	struct fw_cdev_add_descriptor request;
+	int err;
+
+	request.immediate	= immediate_key;
+	request.key		= key;
+	request.data		= ptr_to_u64(data);
+	request.length		= size / 4;
+	request.handle		= 0;
+
+	err = ioctl(handle->local_device->fd, FW_CDEV_IOC_ADD_DESCRIPTOR,
+		    &request);
+	if (err)
+		return err;
+
+	if (token)
+		*token = request.handle;
+
+	return 0;
+}
+
+int
+fw_remove_config_rom_descriptor(fw_handle_t handle, u_int32_t token)
+{
+	struct fw_cdev_remove_descriptor request = {.handle = token};
+
+	return ioctl(handle->local_device->fd, FW_CDEV_IOC_REMOVE_DESCRIPTOR,
+		     &request);
+}
+
 int
 fw_get_config_rom(fw_handle_t handle, quadlet_t *buffer,
 		       size_t buffersize, size_t *rom_size,
@@ -1500,8 +1635,6 @@ fw_get_config_rom(fw_handle_t handle, quadlet_t *buffer,
 	return 0;
 }
 
-#ifdef FW_CDEV_IOC_ALLOCATE_ISO_RESOURCE_ONCE /* added in kernel 2.6.30 */
-
 static int
 iso_resource_modify(raw1394handle_t handle, unsigned int bandwidth,
 		    int channel, enum raw1394_modify_mode mode)
@@ -1547,32 +1680,11 @@ iso_resource_modify(raw1394handle_t handle, unsigned int bandwidth,
 	return 0;
 }
 
-static inline int abi_has_iso_resource_management(raw1394handle_t handle)
-{
-	return handle->mode.fw->abi_version >= 2;
-}
-
-#else
-
-static inline int
-iso_resource_modify(raw1394handle_t handle, unsigned int bandwidth,
-		    int channel, enum raw1394_modify_mode mode)
-{
-	return -1;
-}
-
-static inline int abi_has_iso_resource_management(raw1394handle_t handle)
-{
-	return 0;
-}
-
-#endif  /* defined(FW_CDEV_IOC_ALLOCATE_ISO_RESOURCE_ONCE) */
-
 int
 fw_bandwidth_modify(raw1394handle_t handle, unsigned int bandwidth,
 		    enum raw1394_modify_mode mode)
 {
-	if (abi_has_iso_resource_management(handle))
+	if (handle->mode.fw->abi_version >= 2)
 		return iso_resource_modify(handle, bandwidth, -1, mode);
 	else
 		return ieee1394_bandwidth_modify(handle, bandwidth, mode);
@@ -1582,7 +1694,7 @@ int
 fw_channel_modify(raw1394handle_t handle, unsigned int channel,
 		  enum raw1394_modify_mode mode)
 {
-	if (abi_has_iso_resource_management(handle))
+	if (handle->mode.fw->abi_version >= 2)
 		return iso_resource_modify(handle, 0, channel, mode);
 	else
 		return ieee1394_channel_modify(handle, channel, mode);
