@@ -12,6 +12,7 @@
 #ifndef LIBRAW1394_FW_H
 #define LIBRAW1394_FW_H
 
+#include <stdint.h>
 #include <stdlib.h>
 #include <byteswap.h>
 
@@ -48,7 +49,7 @@ struct fw_handle;
 
 struct epoll_closure {
 	int (*func)(raw1394handle_t handle,
-		    struct epoll_closure *closure, __uint32_t events);
+		    struct epoll_closure *closure, uint32_t events);
 };
 
 struct port {
@@ -131,7 +132,7 @@ struct fw_handle {
 		struct fw_cdev_iso_packet *packets;
 	} iso;
 
-	char buffer[BUFFER_SIZE];
+	__u64 buffer[BUFFER_SIZE / sizeof(__u64)];
 };
 typedef struct fw_handle *fw_handle_t;
 
@@ -144,6 +145,7 @@ int fw_get_fd(fw_handle_t handle);
 nodeid_t fw_get_local_id(fw_handle_t handle);
 nodeid_t fw_get_irm_id(fw_handle_t handle);
 int fw_get_nodecount(fw_handle_t handle);
+int fw_get_speed(fw_handle_t handle, nodeid_t node);
 int fw_get_port_info(fw_handle_t handle, struct raw1394_portinfo *pinf,
 		   int maxports);
 int fw_set_port(fw_handle_t handle, int port);
@@ -202,6 +204,10 @@ int fw_start_fcp_listen(fw_handle_t handle);
 int fw_stop_fcp_listen(fw_handle_t handle);
 int fw_update_config_rom(fw_handle_t handle, const quadlet_t *new_rom,
 		   size_t size, unsigned char rom_version);
+int fw_add_config_rom_descriptor(fw_handle_t handle, u_int32_t *token,
+		   quadlet_t immediate_key, quadlet_t key,
+		   const quadlet_t *data, size_t size);
+int fw_remove_config_rom_descriptor(fw_handle_t handle, u_int32_t token);
 int fw_get_config_rom(fw_handle_t handle, quadlet_t *buffer,
 		   size_t buffersize, size_t *rom_size,
 		   unsigned char *rom_version);
@@ -247,7 +253,9 @@ int fw_iso_recv_unlisten_channel(fw_handle_t handle,
 int fw_iso_recv_set_channel_mask(fw_handle_t handle, u_int64_t mask);
 void fw_iso_stop(fw_handle_t handle);
 void fw_iso_shutdown(fw_handle_t handle);
-int fw_read_cycle_timer(fw_handle_t handle,
-			u_int32_t *cycle_timer, u_int64_t *local_time);
+int fw_read_cycle_timer(fw_handle_t handle, u_int32_t *cycle_timer,
+			  u_int64_t *local_time);
+int fw_read_cycle_timer_and_clock(fw_handle_t handle, u_int32_t *cycle_timer,
+			  u_int64_t *local_time, clockid_t clk_id);
 
 #endif
